from typing import List, Tuple

import bpy
from bpy.types import Object
import mathutils
import math

from ..classes.MdxParts import WarCraft3Model
from ..properties import MDXImportProperties





def scaleVertex(vertex:Tuple[float,float,float], scale:float)->Tuple[float,float,float]:
    return (vertex[0]*scale,vertex[1]*scale,vertex[2]*scale)

def create_armature_object(model: WarCraft3Model, bpy_objects: List[Object], import_properties: MDXImportProperties):
    scale = import_properties.scale_global
    boneSize = import_properties.bone_size*scale
    print("creating armature")
    nodes = model.nodes
    pivot_points = model.pivot_points
    bpy_armature = bpy.data.armatures.new(model.name + ' Nodes')
    bpy_armature.display_type = 'STICK'
    bpy_object = bpy.data.objects.new(model.name + ' Nodes', bpy_armature)
    bpy_object.show_in_front = True
    bpy.context.scene.collection.objects.link(bpy_object)
    bpy_object.select_set(True)
    bpy.context.view_layer.objects.active = bpy_object
    # bpy.context.scene.objects.active = bpy_object
    bpy.ops.object.mode_set(mode='EDIT')

    node_types = set()
    bone_types = {}

    for indexNode, node in enumerate(nodes):
        node_position = pivot_points[indexNode]
        bone_name = node.node.name
        #print(bone_name)
        #temp = BoneRenameMapping.get(bone_name)
        #if temp:
        #    bone_name = temp
        #    node.node.name = temp


        node_types.add(node.type)
        bone = bpy_armature.edit_bones.new(bone_name)
        bone.head = scaleVertex(node_position,scale)
        bone.tail = bone.head
        bone.tail[1] += boneSize
        if bone_name in bone_types.keys():
            bone_name = bone_name + ".001"
            if bone_name in bone_types.keys():
                bone_name = bone_name + ".002"
            node.node.name = bone_name
        bone_types[bone_name] = node.type

    node_types = list(node_types)
    node_types.sort()

    for indexNode, node in enumerate(nodes):
        bone = bpy_armature.edit_bones[indexNode]
        if node.node.parent is not None:
            parent = bpy_armature.edit_bones[node.node.parent]
            bone.parent = parent
            print("node "+bone.name+" parent: {}".format(parent.parent))
            # bone.use_connect = True

    for mesh in bpy_objects:
        mesh.modifiers.new(name='Armature', type='ARMATURE')
        mesh.modifiers['Armature'].object = bpy_object
        for vertexGroup in mesh.vertex_groups:
            vertex_group_index = int(vertexGroup.name)
            bone_name = bpy_armature.edit_bones[vertex_group_index].name
            vertexGroup.name = bone_name
        mesh.parent = bpy_object

    bpy.ops.object.mode_set(mode='POSE')
    bone_groups = {}

    for nodeType in node_types:
        bpy.ops.pose.group_add()
        bone_group = bpy_object.pose.bone_groups.active
        bone_group.name = nodeType + 's'
        bone_groups[nodeType] = bone_group
        if nodeType == 'bone':
            bone_group.color_set = 'THEME04'
        elif nodeType == 'attachment':
            bone_group.color_set = 'THEME09'
        elif nodeType == 'collision_shape':
            bone_group.color_set = 'THEME02'
        elif nodeType == 'event':
            bone_group.color_set = 'THEME03'
        elif nodeType == 'helper':
            bone_group.color_set = 'THEME01'

    for bone in bpy_object.pose.bones:
        #bone.rotation_mode = 'XYZ'
        bone.rotation_mode = 'QUATERNION'
        bone.bone_group = bone_groups[bone_types[bone.name]]

    for bone in bpy_armature.bones:
        bone.warcraft_to_ue.nodeType = bone_types[bone.name].upper()
    
    bpy.ops.object.mode_set(mode='OBJECT')
    bpy.context.active_object.select_set(False)

    return bpy_object
