from typing import Tuple
import bpy, math, mathutils
from bpy.types import Armature, BoneGroup, EditBone, FCurve, Keyframe, NlaStrip, NlaTrack, Object, PoseBone

BoneRenameMapping = {
    "pelvis_bind_jnt":"pelvis",
    "spine_02_bind_jnt":"spine_01",
    "spine_03_bind_jnt":"spine_02",
    "bone_chest":"spine_03",
    "L_shoulder_bind_jnt":"clavicle_l",
    "R_shoulder_bind_jnt":"clavicle_r",
    "L_breast_bind_jnt":"breast_l",
    "R_breast_bind_jnt":"breast_r",
    "neck_bind_jnt":"neck_01",
    "bone_head":"head",
    
    "L_upr_arm_bind_jnt":"upperarm_l",
    "L_lwr_arm_bind_jnt":"lowerarm_l",
    "bone_hand_left":"hand_l",
    "L_lwr_arm_twist_01_bind_jnt":"lowerarm_twist_01_l",
    "L_lwr_arm_twist_02_bind_jnt":"lowerarm_twist_02_l",

    "R_upr_arm_bind_jnt":"upperarm_r",
    "R_lwr_arm_bind_jnt":"lowerarm_r",
    "bone_hand_right":"hand_r",
    "R_lwr_arm_twist_01_bind_jnt":"lowerarm_twist_01_r",
    "R_lwr_arm_twist_02_bind_jnt":"lowerarm_twist_02_r",

    "L_leg_01_bind_jnt":"thigh_l",
    "L_leg_02_bind_jnt":"calf_l",
    "bone_leg_left":"foot_l",
    "L_leg_04_bind_jnt":"ball_l",
    
    "R_leg_01_bind_jnt":"thigh_r",
    "R_leg_02_bind_jnt":"calf_r",
    "bone_leg_right":"foot_r",
    "R_leg_04_bind_jnt":"ball_r",
    
    "weapon_bind_jnt":"weapon",


    "L_ind_finger_01_bind_jnt":"index_01_l",
    "L_ind_finger_02_bind_jnt":"index_02_l",
    "L_ind_finger_03_bind_jnt":"index_03_l",

    "L_mid_finger_01_bind_jnt":"middle_01_l",
    "L_mid_finger_02_bind_jnt":"middle_02_l",
    "L_mid_finger_03_bind_jnt":"middle_03_l",

    "L_pnk_finger_01_bind_jnt":"pinky_01_l",
    "L_pnk_finger_02_bind_jnt":"pinky_02_l",
    "L_pnk_finger_03_bind_jnt":"pinky_03_l",

    "L_rng_finger_01_bind_jnt":"ring_01_l",
    "L_rng_finger_02_bind_jnt":"ring_02_l",
    "L_rng_finger_03_bind_jnt":"ring_03_l",
    
    "L_tmb_01_bind_jnt":"thumb_01_l",
    "L_tmb_02_bind_jnt":"thumb_02_l",
    "L_tmb_03_bind_jnt":"thumb_03_l",

    "R_ind_finger_01_bind_jnt":"index_01_r",
    "R_ind_finger_02_bind_jnt":"index_02_r",
    "R_ind_finger_03_bind_jnt":"index_03_r",

    "R_mid_finger_01_bind_jnt":"middle_01_r",
    "R_mid_finger_02_bind_jnt":"middle_02_r",
    "R_mid_finger_03_bind_jnt":"middle_03_r",

    "R_pnk_finger_01_bind_jnt":"pinky_01_r",
    "R_pnk_finger_02_bind_jnt":"pinky_02_r",
    "R_pnk_finger_03_bind_jnt":"pinky_03_r",

    "R_rng_finger_01_bind_jnt":"ring_01_r",
    "R_rng_finger_02_bind_jnt":"ring_02_r",
    "R_rng_finger_03_bind_jnt":"ring_03_r",

    "R_tmb_01_bind_jnt":"thumb_01_r",
    "R_tmb_02_bind_jnt":"thumb_02_r",
    "R_tmb_03_bind_jnt":"thumb_03_r",

}

BoneReparentFixes=[
    # bone                             ,  parent              , reparent ,invert, clr  , offset
    ('bone_turret'                     , 'pelvis'             , '.'      , True , True , False),
    (('spine_01','spine_02','spine_03'), 'bone_turret'        , 'pelvis' , False, False, True ),
    ('hand_l'                          , 'lowerarm_twist_02_l', '.parent', False, False, True),
    ('hand_r'                          , 'lowerarm_twist_02_r', '.parent', False, False, True),

]

ROTATION_EULER = "pose.bones[\"{}\"].rotation_euler"
ROTATION_QUAT = "pose.bones[\"{}\"].rotation_quaternion"
LOCATION = "pose.bones[\"{}\"].location"

def recalculate_bone(obj:Object ,strip:NlaStrip,target:EditBone,parent:EditBone,inverse:bool=False,fix_offset:bool=False):
    print("  - recalculating ",target.name,parent.name," in ",strip.name)
    poseTarget:PoseBone = obj.pose.bones.get(target.name)
    poseParent:PoseBone = obj.pose.bones.get(parent.name)
    pathTarget = ROTATION_QUAT.format(target.name)
    pathParent = ROTATION_QUAT.format(parent.name)
    pathTargetLoc = LOCATION.format(target.name)
    tcs = ( strip.action.fcurves.find(pathTarget,index=0), strip.action.fcurves.find(pathTarget,index=1), strip.action.fcurves.find(pathTarget,index=2), strip.action.fcurves.find(pathTarget,index=3))
    pcs = ( strip.action.fcurves.find(pathParent,index=0), strip.action.fcurves.find(pathParent,index=1), strip.action.fcurves.find(pathParent,index=2), strip.action.fcurves.find(pathParent,index=3))
    #strip.action.keyframe_insert(pathTargetLoc,-1,0,"",)
    #target.keyframe_insert('location',-1,0)

    if fix_offset:
        pos = (strip.action.fcurves.find(pathTargetLoc,index=0), strip.action.fcurves.find(pathTargetLoc,index=1), strip.action.fcurves.find(pathTargetLoc,index=2))
        for i in range(3):
            if not pos[i]:
                fcurve = strip.action.fcurves.new(pathTargetLoc, index=i, action_group=target.name)
                fcurve.keyframe_points.insert(0.0, 0.0)
        pos = (strip.action.fcurves.find(pathTargetLoc,index=0), strip.action.fcurves.find(pathTargetLoc,index=1), strip.action.fcurves.find(pathTargetLoc,index=2))

        
            
        

        th = mathutils.Vector(target.head)
        ph = mathutils.Vector(parent.head)
        d:mathutils.Vector = th-ph
    
    #print(d)
    if not tcs[0] or not tcs[1] or not tcs[2] or not tcs[3]:
        return
    if not pcs[0] or not pcs[1] or not pcs[2] or not pcs[3]:
        return
    if inverse:
        for i in range(0,tcs[0].keyframe_points.__len__()):
            time = tcs[0].keyframe_points[i].co[0]
            tq:mathutils.Quaternion = mathutils.Quaternion((tcs[0].keyframe_points[i].co[1],tcs[1].keyframe_points[i].co[1],tcs[2].keyframe_points[i].co[1],tcs[3].keyframe_points[i].co[1]))
            pq:mathutils.Quaternion = mathutils.Quaternion((pcs[0].evaluate(time),pcs[1].evaluate(time),pcs[2].evaluate(time),pcs[3].evaluate(time)))
            pq.invert()
            tq.rotate(pq)

            tcs[0].keyframe_points[i].co[1] = tq.w
            tcs[1].keyframe_points[i].co[1] = tq.x
            tcs[2].keyframe_points[i].co[1] = tq.y
            tcs[3].keyframe_points[i].co[1] = tq.z
            #a:mathutils.Euler = mathutils.Euler((tcs[0].keyframe_points[i].co[1],tcs[1].keyframe_points[i].co[1],tcs[2].keyframe_points[i].co[1]), 'XYZ')
            #b:mathutils.Euler = mathutils.Euler((pcs[0].evaluate(time),pcs[1].evaluate(time),pcs[2].evaluate(time)), 'XYZ')
            #c:mathutils.Quaternion = b.to_quaternion()
            #c.invert()
            #a.rotate(c)
            #tcs[0].keyframe_points[i].co[1] = a.x
            #tcs[1].keyframe_points[i].co[1] = a.y
            #tcs[2].keyframe_points[i].co[1] = a.z
    else:
        if fix_offset:
            for i in range(0,tcs[0].keyframe_points.__len__()):
                time = tcs[0].keyframe_points[i].co[0]
                tq:mathutils.Quaternion = mathutils.Quaternion((tcs[0].keyframe_points[i].co[1],tcs[1].keyframe_points[i].co[1],tcs[2].keyframe_points[i].co[1],tcs[3].keyframe_points[i].co[1]))
                pq:mathutils.Quaternion = mathutils.Quaternion((pcs[0].evaluate(time),pcs[1].evaluate(time),pcs[2].evaluate(time),pcs[3].evaluate(time)))
                tq.rotate(pq)
                tcs[0].keyframe_points[i].co[1] = tq.w
                tcs[1].keyframe_points[i].co[1] = tq.x
                tcs[2].keyframe_points[i].co[1] = tq.y
                tcs[3].keyframe_points[i].co[1] = tq.z
                
                tmp = d.copy()
                tmp.rotate(pq)
                n:mathutils.Vector = tmp-d
                keyframe = pos[0].keyframe_points.insert(time, n.x)
                keyframe = pos[1].keyframe_points.insert(time, n.y)
                keyframe = pos[2].keyframe_points.insert(time, n.z)
        else:
            for i in range(0,tcs[0].keyframe_points.__len__()):
                time = tcs[0].keyframe_points[i].co[0]
                tq:mathutils.Quaternion = mathutils.Quaternion((tcs[0].keyframe_points[i].co[1],tcs[1].keyframe_points[i].co[1],tcs[2].keyframe_points[i].co[1],tcs[3].keyframe_points[i].co[1]))
                pq:mathutils.Quaternion = mathutils.Quaternion((pcs[0].evaluate(time),pcs[1].evaluate(time),pcs[2].evaluate(time),pcs[3].evaluate(time)))
                tq.rotate(pq)
                tcs[0].keyframe_points[i].co[1] = tq.w
                tcs[1].keyframe_points[i].co[1] = tq.x
                tcs[2].keyframe_points[i].co[1] = tq.y
                tcs[3].keyframe_points[i].co[1] = tq.z

            #a:mathutils.Euler = mathutils.Euler((tcs[0].keyframe_points[i].co[1],tcs[1].keyframe_points[i].co[1],tcs[2].keyframe_points[i].co[1]), 'XYZ')
            #b:mathutils.Euler = mathutils.Euler((pcs[0].evaluate(time),pcs[1].evaluate(time),pcs[2].evaluate(time)), 'XYZ')
            #c:mathutils.Quaternion = b.to_quaternion()
            #a.rotate(c)
            #tcs[0].keyframe_points[i].co[1] = a.x
            #tcs[1].keyframe_points[i].co[1] = a.y
            #tcs[2].keyframe_points[i].co[1] = a.z

def fix_character_model(context):
    print("Fixing character model:")
    ctx:bpy.context = bpy.context.copy()

    
    #for key, value in ctx.items():
    #    print(key)
    #    print(value)

    #print(ctx)

    
    #ctx['area'].type = 'VIEW_3D'
    obj:bpy.context.object = ctx['object']

    if obj and obj.type == 'ARMATURE':
        arm:Armature = obj.data

        bpy.ops.object.mode_set(ctx, mode='EDIT')
        try:

            pbone:PoseBone
            for pbone, bone in zip(obj.pose.bones, arm.edit_bones):
                if pbone.bone_group.name == 'collision_shapes' or pbone.bone_group.name=='events':
                    arm.edit_bones.remove(bone)

            for bone in arm.edit_bones:
                temp = BoneRenameMapping.get(bone.name)
                if temp:
                    bone.name = temp




            #for gr in obj.pose.bone_groups:
            #    gr.

            recalcList = []
            clenupList = []
            reparentList = []

            for dat in BoneReparentFixes:
                target:EditBone
                reparent:EditBone
                parent:EditBone = arm.edit_bones.get(dat[1])
                if not parent:
                    continue
                if dat[2]=='.':
                    reparent = parent
                elif dat[2]=='.parent':
                    reparent = parent.parent
                else:
                    reparent = arm.edit_bones.get(dat[2])
                if type(dat[0]) is tuple:
                    for x in dat[0]:
                        target = arm.edit_bones.get(x)
                        if target:
                            break
                else:
                    target = arm.edit_bones.get(dat[0])
                if not target:
                    continue
                recalcList.append((target, parent, dat[3], dat[5]))
                reparentList.append((target, reparent))
                if dat[4]:
                    clenupList.append((target,))

            if obj.animation_data:
                t:NlaTrack
                for t in obj.animation_data.nla_tracks:
                    if t.name.startswith("Decay"):
                        for s in t.strips:
                            #t.strips.remove(s)
                            bpy.data.actions.remove(s.action)
                        obj.animation_data.nla_tracks.remove(t)
                        continue

                    s:NlaStrip
                    for s in t.strips:
                        fix:Tuple[EditBone,EditBone,bool,bool]
                        for fix in recalcList:
                            tn = fix[0].name
                            pn = fix[1].name
                            #recalculate_bone(s,ROTATION_EULER.format(tn),ROTATION_EULER.format(pn),fix[2])
                            recalculate_bone(obj,s,fix[0],fix[1],fix[2],fix[3])

                        fix:Tuple[EditBone]
                        for fix in clenupList:
                            bone:EditBone = fix[0]
                            n = bone.name
                            locpath = LOCATION.format(n)
                            rotpath = ROTATION_EULER.format(n)
                            rotpathq = ROTATION_QUAT.format(n)
                            for i in (0,1,2):
                                tmp = s.action.fcurves.find(locpath,index=i)
                                if tmp:
                                    s.action.fcurves.remove(tmp)
                                tmp = s.action.fcurves.find(rotpath,index=i)
                                if tmp:
                                    s.action.fcurves.remove(tmp)
                            for i in range(4):
                                tmp = s.action.fcurves.find(rotpathq,index=i)
                                if tmp:
                                    s.action.fcurves.remove(tmp)

            fix:Tuple[EditBone,EditBone]
            for fix in reparentList:
                bone:EditBone = fix[0]
                rep = fix[1]
                bone.parent = rep
                #print(" reparenting ", bone, rep)
            
            fix:Tuple[EditBone]
            for fix in clenupList:
                fix[0].use_deform = False
            
        finally:
            bpy.ops.object.mode_set(ctx, mode='OBJECT')
            bpy.ops.transform.rotate(ctx, orient_axis='Z',value=math.radians(90))
            bpy.ops.object.transform_apply(ctx, location=False, rotation=True, scale=False)
    print("Done!")

