from types import FunctionType
from typing import Tuple
from ..classes.MdxParts import (
    FaceFxEffect,
    WarCraft3MeshGeo,
    WarCraft3Material,
    WarCraft3Texture,
    WarCraft3Bone,
    WarCraft3GeosetTransformation,
    WarCraft3Node,
    WarCraft3Layer,
    WarCraft3Helper,
    WarCraft3Attachment,
    WarCraft3CollisionShape,
    WarCraft3Event,
    WarCraft3GeosetAnimation,
    WarCraft3Sequence,
    WarCraft3Model
)
from .. import constants
from .binary_reader import StreamReader

def readVersion(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()
    version = reader.readInt()

    print("mdl version: ", version)
    if version in constants.MDX_VERSIONS:
        model.version = version
        constants.MDX_CURRENT_VERSION = version
    else:
        print("Version %s is not supported; the model will load as %s which might cause issues"
              % (version, model.version))
        # raise Exception('unsupported MDX format version: {0}'.format(version))


def readModelHeader(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()
    model.name = reader.readStr(80)
    animation_file_name = reader.readStr(260)
    bounds_radius = reader.readF('<f')[0]
    minimum_extent = reader.readF('<3f')
    maximum_extent = reader.readF('<3f')
    blend_time = reader.readInt()


def readSequences(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()
    if size % 132 != 0:
        raise Exception('bad sequence data (size % 132 != 0)')
    sequence_count = size // 132
    for _ in range(sequence_count):
        sequence = WarCraft3Sequence()
        sequence.name = reader.readStr(80)
        sequence.interval_start = reader.readSInt32()
        sequence.interval_end = reader.readSInt32()
        sequence.movement_speed = reader.readFloat()
        sequence.flags = reader.readSInt32()
        sequence.rarity = reader.readFloat()
        sequence.sync_point = reader.readFloat()
        sequence.bounds_radius = reader.readFloat()
        sequence.minimum_extent = reader.readVertex()
        sequence.maximum_extent = reader.readVertex()
        model.sequences.append(sequence)

def readGlobalSequences(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()
    model.globalSequences = reader.readUIntArray(size//4)

def readTextures(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()
    if size % 268 != 0:
        raise Exception('bad Texture data (size % 268 != 0)')
    textures_count = size // 268
    for _ in range(textures_count):
        texture = WarCraft3Texture()
        texture.replaceable_id = reader.readInt()
        texture.image_file_name = reader.readStr(260)
        texture.flags = reader.readInt()
        model.textures.append(texture)

def readPivotData(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()

    if size % 12 != 0:
        raise Exception('bad Pivot Point data (size % 12 != 0)')

    pivot_points_count = size // 12

    for _ in range(pivot_points_count):
        model.pivot_points.append(reader.readVertex())

def readFaceFX(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()
    if size % 340 != 0:
        raise Exception('bad FaceFX data (size % 340 != 0)')
    count = size // 340
    for _ in range(count):
        ffx = FaceFxEffect()
        ffx.type = reader.readStr(80)
        ffx.path = reader.readStr(260)
        model.face_effects.append(ffx)


def readBindPose(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()
    for _ in range(reader.readInt()):
        model.bind_pose.append(reader.readF('<12f'))



def readDynamicChunk(reader:StreamReader, model: WarCraft3Model, impl:FunctionType):
    size = reader.readInt32()
    print("---reading dynamic chunk of size {}".format(size))
    while size > 0:
        mark = reader.mark
        object = impl(reader,model)
        if object:
            size -= reader.mark-mark
        #print("---tail left {}".format(size))
    #print("---dynamic chunk complete")


def readGeoset(reader:StreamReader, model: WarCraft3Model):
    size = reader.readInt32()
    #print("reading geoset of size {}".format(size))

    mesh = WarCraft3MeshGeo()

    reader.read4cc() # skip VRTX
    mesh.vertices = reader.readVertexList(reader.readInt())
    reader.read4cc() # skip NRMS
    mesh.normals = reader.readVertexList(reader.readInt())
    reader.read4cc() # skip PTYP
    mesh.faceTypeGroups = reader.readUIntArray(reader.readInt())
    reader.read4cc() # skip PCNT
    mesh.faceGroups = reader.readUIntArray(reader.readInt())
    reader.read4cc() # skip PVTX
    #mesh.faces = reader.readArray(reader.readInt(),'I')
    mesh.faces = reader.readTriangleList(reader.readInt()//3)
    reader.read4cc() # skip GNDX
    mesh.vertexGroups = reader.readArray(reader.readInt(),'B')
    reader.read4cc() # skip MTGC
    mesh.matrixGroups = reader.readUIntArray(reader.readInt())
    reader.read4cc() # skip MATS
    mesh.matrixIndices = reader.readUIntArray(reader.readInt())
    
    

    mesh.materialId = reader.readInt()
    mesh.selectionGroup = reader.readInt()
    mesh.selectionFlags = reader.readInt()

    if (model.version > 800):
        mesh.lod = reader.readInt()
        mesh.lodName = reader.readStr(80)

    mesh.extent = (reader.readFloat(),reader.readVertex(),reader.readVertex())
    mesh.extents_count = reader.readInt()

    for _ in range(mesh.extents_count):
        mesh.extents.append((reader.readFloat(),reader.readVertex(),reader.readVertex()))
    #ToDo: add read extent

    id = reader.read4cc()
    #print(id)
    if model.version > 800 and id != constants.CHUNK_TEXTURE_VERTEX_GROUP:
        if id == constants.CHUNK_TANGENTS:
            mesh.tangents = reader.readFloatArray(reader.readInt() * 4)
            id = reader.read4cc()   # SKIN or UVAS
            #print(id)
        if id == constants.CHUNK_SKIN:
            mesh.skin = reader.readArray(reader.readInt(),'B')
            #print(mesh.skin)
            id = reader.read4cc()   # UVAS
            #print(id)

    numUV = reader.readInt()
    
    #print(id+" reading uvs {} ".format(numUV))
    for i in range(numUV):
        reader.read4cc(); # skip UVBS
        numv = reader.readInt()
        #print("reading uv vertexes {}".format(numv))
        mesh.uvs.append(reader.readVertexList2(numv))

    model.meshes.append(mesh)

    return mesh

def readTimeline(reader:StreamReader, getter, processor=None):
    mark=reader.mark
    transformation = WarCraft3GeosetTransformation()
    transformation.tracks_count = reader.readInt32()
    transformation.interpolation_type = reader.readInt32()
    transformation.global_sequence_id = reader.readSInt32()
    for _ in range(transformation.tracks_count):
        time = reader.readSInt32()
        values = getter()    # translation values
        if processor:
            #print(values)
            values=processor(values)
        transformation.times.append(time)
        transformation.values.append(values)
        if transformation.interpolation_type > constants.INTERPOLATION_TYPE_LINEAR:
            print("nonlinear")
            in_tan = getter()
            out_tan = getter()
            if processor:
                in_tan=processor(in_tan)
                out_tan=processor(out_tan)
            transformation.in_tans.append(in_tan)
            transformation.out_tans.append(out_tan)
    return transformation

def swapRotValues(values:Tuple[float,float,float,float])->Tuple[float,float,float,float]:
    return (values[3], values[0], values[1], values[2])

def readNode(reader:StreamReader)->WarCraft3Node:
    node = WarCraft3Node()
    size = reader.readInt32()
    #print("reading node data of size {}".format(size))
    node.name = reader.readStr(80)
    #print(" node "+node.name+" {}".format(size))
    node.id = reader.readSInt32()
    node.parent = reader.readSInt32()
    if node.parent == -1:
        node.parent = None
    print("node "+node.name+" parent: {}".format(node.parent))
    node.flags = reader.readSInt32()
    #print(" node "+node.name+" {} {} {} {}".format(size,node.id,node.parent,node.flags))


    size -= 96
    mark = reader.mark
    while size > 0:
        chunk_id = reader.read4cc()
        #print("  track "+chunk_id+" {}".format(size-4))
        if chunk_id == constants.CHUNK_GEOSET_TRANSLATION:
            node.translations = readTimeline(reader,reader.readVertex)
        elif chunk_id == constants.CHUNK_GEOSET_ROTATION:
            node.rotations = readTimeline(reader,reader.readVertex4,swapRotValues)
        elif chunk_id == constants.CHUNK_GEOSET_SCALING:
            node.scalings = readTimeline(reader,reader.readVertex)
        else:
            print("unknown timeline type "+chunk_id+" skipping remaining data")
            size -= reader.mark-mark
            reader.skip(size)
            size = 0
            break
        size -= reader.mark-mark
        mark = reader.mark
        #if size!=0:
        #    print("   tail {}".format(size))
    return node

def readBone(reader:StreamReader, model: WarCraft3Model)->WarCraft3Bone:
    #size = reader.readInt()
    #print("reading bone data")

    bone = WarCraft3Bone()
    bone.node = readNode(reader)
    bone.geoset_id = reader.readInt32()
    bone.geoset_animation_id = reader.readInt32()
    model.nodes.append(bone)
    return bone


def readMaterialLayer(reader:StreamReader, model: WarCraft3Model)->WarCraft3Layer:
    layer = WarCraft3Layer()
    mark = reader.mark
    size = reader.readInt32()

    layer.filterMode = reader.readSInt32()
    layer.shadingFlags = reader.readSInt32()
    layer.texture_id = reader.readSInt32()
    layer.textureAnimationId = reader.readSInt32()
    layer.coordId = reader.readSInt32()
    layer.alpha = reader.readFloat()
    # if constants.MDX_CURRENT_VERSION > 800:
    if model.version > 800:
        layer.emissive_gain = reader.readFloat()
        # if constants.MDX_CURRENT_VERSION > 900:
        if model.version > 900:
            layer.fresnel_color = reader.readVertex()#[r.getf('<f')[0], r.getf('<f')[0], r.getf('<f')[0]]
            layer.fresnel_opacity = reader.readFloat()
            layer.fresnel_team_color =reader.readFloat()
    size -= reader.mark-mark
    mark = reader.mark
    while size>0:
        chunk_id = reader.read4cc()
        if chunk_id == constants.CHUNK_MATERIAL_TEXTURE_ID:
            layer.material_texture_id = readTimeline(reader,reader.readSInt32)
        elif chunk_id == constants.CHUNK_MATERIAL_ALPHA:
            layer.material_alpha = readTimeline(reader,reader.readFloat)
        elif chunk_id == constants.CHUNK_MATERIAL_FRESNEL_COLOR:
            layer.fresnel_color = readTimeline(reader,reader.readVertex)
        elif chunk_id == constants.CHUNK_MATERIAL_EMISSIONS:
            layer.emissions = readTimeline(reader,reader.readFloat)
        elif chunk_id == constants.CHUNK_MATERIAL_FRESNEL_ALPHA:
            layer.fresnel_alpha = readTimeline(reader,reader.readFloat)
        elif chunk_id == constants.CHUNK_MATERIAL_FRESNEL_TEAMCOLOR:
            layer.fresnel_team_color = readTimeline(reader,reader.readInt32)
        else:
            print("unknown layer subchunk, skipping remaining")
            size -= reader.mark-mark
            reader.skip(size)
            size = 0
            break
        size -= reader.mark-mark
        mark = reader.mark

    return layer

def readMaterial(reader:StreamReader, model: WarCraft3Model)->WarCraft3Material:
    material = WarCraft3Material()
    size = reader.readInt32()
    priorityPlane = reader.readSInt32()
    flags = reader.readSInt32()

    if model.version > 800:
        shader = reader.readStr(80)
        if shader == "Shader_HD_DefaultUnit":
            material.hd = True

    reader.read4cc() # skip LAYS

    layerCount = reader.readInt32()
    for _ in range(layerCount):
        layer:WarCraft3Layer = readMaterialLayer(reader, model)
        material.layers.append(layer)
    model.materials.append(material)

    return material

def readHelper(reader:StreamReader, model: WarCraft3Model)->WarCraft3Helper:
    helper = WarCraft3Helper()
    helper.node = readNode(reader)
    model.nodes.append(helper)
    return helper

def readAttachment(reader:StreamReader, model: WarCraft3Model)->WarCraft3Attachment:
    mark = reader.mark
    size = reader.readInt32()
    attachment = WarCraft3Attachment()
    attachment.node = readNode(reader)
    attachment.path = reader.readStr(260)
    attachment.attachment_id = reader.readSInt32()

    size -= reader.mark-mark
    mark = reader.mark
    while size>0:
        chunk_id = reader.read4cc()
        if chunk_id == constants.CHUNK_ATTACHMENT_VISIBILITY:
            attachment.visibility = readTimeline(reader,reader.readFloat)
        else:
            print("unknown attachment subchunk, skipping remaining")
            size -= reader.mark-mark
            reader.skip(size)
            size = 0
            break
        size -= reader.mark-mark
        mark = reader.mark

    model.nodes.append(attachment)
    return attachment


def readEvent(reader:StreamReader, model: WarCraft3Model)->WarCraft3Event:
    event = WarCraft3Event()
    event.node = readNode(reader)
    reader.read4cc() # KEVT skipped
    tracks_count = reader.readInt32()
    event.global_sequence_id = reader.readSInt32()
    event.keys = tracks_count
    event.timeline=reader.readUIntArray(tracks_count)
    model.nodes.append(event)
    return event

def readCollisionShape(reader:StreamReader, model: WarCraft3Model)->WarCraft3CollisionShape:
    collision_shape = WarCraft3CollisionShape()
    collision_shape.node = readNode(reader)
    collision_type = reader.readSInt32()
    collision_shape.shape = collision_type
    if collision_type == 0 or collision_type == 1 or collision_type == 3:
        vertices_count = 2
    elif collision_type == 2:
        vertices_count = 1
    else:
        raise Exception('UNSUPPORTED COLLISION SHAPE TYPE:', collision_type)

    collision_shape.points = reader.readVertexList(vertices_count)

    if collision_type == 2 or collision_type == 3:
        collision_shape.radius = reader.readFloat()

    model.nodes.append(collision_shape)
    return collision_shape

def readGeosetAnimation(reader:StreamReader, model: WarCraft3Model)->WarCraft3GeosetAnimation:
    geoset_animation = WarCraft3GeosetAnimation()
    mark = reader.mark
    size = reader.readInt32()
    alpha = reader.readFloat()
    geoset_animation.flags = reader.readSInt32()
    color = reader.readVertex()
    geoset_animation.geoset_id = reader.readSInt32()

    size -= reader.mark-mark
    mark = reader.mark
    while size>0:
        chunk_id = reader.read4cc()
        if chunk_id == constants.CHUNK_GEOSET_COLOR:
            geoset_animation.animation_color = readTimeline(reader,reader.readVertex)
        elif chunk_id == constants.CHUNK_GEOSET_ALPHA:
            geoset_animation.animation_alpha = readTimeline(reader,reader.readFloat)
        else:
            print("unknown attachment subchunk, skipping remaining")
            size -= reader.mark-mark
            reader.skip(size)
            size = 0
            break
        size -= reader.mark-mark
        mark = reader.mark

    if not geoset_animation.animation_color:
        geoset_color = WarCraft3GeosetTransformation()
        geoset_color.tracks_count = 1
        geoset_color.interpolation_type = constants.INTERPOLATION_TYPE_NONE
        geoset_color.times = [0, ]
        geoset_color.values = [color, ]
        geoset_animation.animation_color = geoset_color

    if not geoset_animation.animation_alpha:
        geoset_alpha = WarCraft3GeosetTransformation()
        geoset_alpha.tracks_count = 1
        geoset_alpha.interpolation_type = constants.INTERPOLATION_TYPE_NONE
        geoset_alpha.times = [0, ]
        geoset_alpha.values = [alpha, ]
        geoset_animation.animation_alpha = geoset_alpha

    model.geoset_animations.append(geoset_animation)
    return geoset_animation